function DFST
% Integrating DFST tracker into the VOT toolkit - VOT 2016
% Dynamic Feature Selection tracker VOT integration
% Giorgio Roffo - giorgio.roffo@univr.it

% *************************************************************
% VOT: Always call exit command at the end to terminate Matlab!
% *************************************************************
cleanup = onCleanup(@() exit() );

% *************************************************************
% VOT: Set random seed to a different value every time.
% *************************************************************
RandStream.setGlobalStream(RandStream('mt19937ar', 'Seed', sum(clock)));

% **********************************
% VOT: Get initialization data
% **********************************
[handle, image, region] = vot('rectangle');

% Initialize the tracker
[state, ~] = dfst_initialize(imread(image), region);

while true
    
    % **********************************
    % VOT: Get next frame
    % **********************************
    [handle, image] = handle.frame(handle);
    
    if isempty(image)
        break;
    end;
    
    % Perform a tracking step, obtain new region
    [state, region] = dfst_update(state, imread(image));
    
    % **********************************
    % VOT: Report position for frame
    % **********************************
    handle = handle.report(handle, region);
    
end;

% **********************************
% VOT: Output the results
% **********************************
handle.quit(handle);

end

function state = dfst_create_state(region)

state.padding = 1.0;         			   % extra area surrounding the target
state.output_sigma_factor = 1/16;		   % spatial bandwidth (proportional to target)
state.sigma = 0.2;                         % gaussian kernel bandwidth
state.lambda = 1e-2;					   % regularization
state.learning_rate = 0.005;               % learning rate for appearance model update scheme
state.compression_learning_rate = 0.1;     % learning rate for the adaptive dimensionality reduction
state.non_compressed_features = {'gray'};  % features that are not compressed, a cell with strings (possible choices: 'gray', 'cn')
state.compressed_features = {'cn'};        % features that are compressed, a cell with strings (possible choices: 'gray', 'cn')
state.num_selected_dim = 8;                % number of features to select 8
state.num_compressed_dim = 4;              % the dimensionality of the compressed features 2
state.K=250;                               % learns a dictionary with 200 elements
state.lambda=0.15;
state.numThreads=4;                        % number of threads
state.batchsize=100;
state.verbose=false;
state.iter=200;
% If the provided region is a polygon ...
if numel(region) > 4
    x1 = round(min(region(1:2:end)));
    x2 = round(max(region(1:2:end)));
    y1 = round(min(region(2:2:end)));
    y2 = round(max(region(2:2:end)));
    region = round([x1, y1, x2 - x1, y2 - y1]);
else
    region = round([round(region(1)), round(region(2)), ...
        round(region(1) + region(3)) - round(region(1)), ...
        round(region(2) + region(4)) - round(region(2))]);
end;
state.wsize = floor([ region(1,4), region(1,3)]);
state.init_pos = floor([region(1,2), region(1,1)]) + floor(state.wsize/2);
% load the normalized Color Name matrix - (Cordelia Schmid - Color Names)
temp = load('w2crs');
state.w2c = temp.w2crs;
end

%% Dynamic Feature Selection tracker: Initialization
function [state, location] = dfst_initialize(im, region, varargin)

% Create state for DFST
state = dfst_create_state(region);

% parameters
output_sigma_factor = state.output_sigma_factor;
sigma = state.sigma;
lambda = state.lambda;
non_compressed_features = state.non_compressed_features;
compressed_features = state.compressed_features;
num_selected_dim = state.num_selected_dim;
num_compressed_dim = state.num_compressed_dim ;
pos = floor(state.init_pos);
target_sz = floor(state.wsize);
w2c = state.w2c; % Color Name matrix

[rows, cls, channels] = size(im);
% RGB images
if channels==1
    im = cat(3,im,im,im);
end

% window size, taking padding into account
padding_s1 = min(round((rows/target_sz(1))/2), 4);
padding_s2 = min(round((cls/target_sz(2))/2), 4);
state.sz(1) = floor(target_sz(1) * padding_s1);
state.sz(2) = floor(target_sz(2) * padding_s2);
% state.sz =  floor(target_sz * (1 + padding));

rect_bbox = [pos(2)-target_sz(2)/2,  pos(1)-target_sz(1)/2, target_sz(2) target_sz(1)];
% imshow(sub_patch)
I = imresize(double(imcrop(im,rect_bbox))/255,[state.wsize(1),state.wsize(2)]);
% extract 8 x 8 patches
X = im2col(I,[4 4],'sliding');
X = X - repmat(mean(X),[size(X,1) 1]);
X = X ./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
state.D = mexTrainDL(X,state);
state.alpha = mexLasso(X,state.D,state);

% desired output (gaussian shaped), bandwidth proportional to target size
output_sigma = sqrt(prod(target_sz)) * output_sigma_factor;
[rs, cs] = ndgrid((1:state.sz(1)) - floor(state.sz(1)/2), (1:state.sz(2)) - floor(state.sz(2)/2));
y = exp(-0.5 / output_sigma^2 * (rs.^2 + cs.^2));
state.yf = single(fft2(y));

% store pre-computed cosine window
state.cos_window = single(hann(state.sz(1)) * hann(state.sz(2))');

% extract the feature map of the local image patch to train the classifer
[xo_npca, xo_pca, xo_pca_neg ] = get_subwindow_fs(im, pos, state.sz, non_compressed_features, compressed_features, w2c);
state.z_npca = xo_npca;
state.z_pca = xo_pca;

% FEATURE SELECTION
[ feature_ranking ] = supINFFS( xo_pca_neg, xo_pca );

% compute the mean appearance
data_mean = mean(state.z_pca, 1);

% substract the mean from the appearance to get the data matrix
data_matrix = bsxfun(@minus, state.z_pca, data_mean);

data_matrix(:,feature_ranking(num_selected_dim+1:end))=0;
data_matrix = data_matrix(:,  feature_ranking);
% calculate the covariance matrix
cov_matrix = 1/(prod(state.sz) - 1) * (data_matrix' * data_matrix);

% calculate the principal components (pca_basis) and corresponding variances
[pca_basis, pca_variances, ~] = svd(cov_matrix);

% calculate the projection matrix as the first principal
% components and extract their corresponding variances
state.projection_matrix = [pca_basis(:, 1:num_compressed_dim)];
state.projection_variances = pca_variances(1:num_compressed_dim, 1:num_compressed_dim);

% initialize the old covariance matrix using the computed
% projection matrix and variances
state.old_cov_matrix = state.projection_matrix * state.projection_variances * state.projection_matrix';

% project the features of the new appearance example using the new
% projection matrix
x = feature_projection(xo_npca, xo_pca, state.projection_matrix, state.cos_window);

% calculate the new classifier coefficients
kf = fft2(dense_gauss_kernel(sigma, x));
new_alphaf_num = state.yf .* kf;
new_alphaf_den = kf .* (kf + lambda);

% first frame, train with a single image
state.alphaf_num = new_alphaf_num;
state.alphaf_den = new_alphaf_den;

location = [ [pos([2,1]) - (target_sz([2,1])-1)/2 ], target_sz([2,1])];

state.position = pos;
state.newsize = target_sz;
end


%% Dynamic Feature Selection tracker: Update Function
function [state, location] = dfst_update(state, im, varargin)

% read parameters
sigma = state.sigma;
lambda = state.lambda;
learning_rate = state.learning_rate;
compression_learning_rate = state.compression_learning_rate;
non_compressed_features = state.non_compressed_features;
compressed_features = state.compressed_features;
num_selected_dim = state.num_selected_dim;
num_compressed_dim = state.num_compressed_dim ;
target_sz = floor(state.newsize);
w2c = state.w2c; % Color Name matrix
sz = state.sz;

pos = floor(state.position); % last position

[rows,cls,channels] = size(im);
if channels==1
    im = cat(3,im,im,im);
end

% compute the compressed learnt appearance
zp = feature_projection(state.z_npca, state.z_pca, state.projection_matrix, state.cos_window);

% extract the feature map of the local image patch
[xo_npca, xo_pca] = get_subwindow_fast(im, pos, sz, non_compressed_features, compressed_features, w2c);

% do the dimensionality reduction and windowing
x = feature_projection(xo_npca, xo_pca, state.projection_matrix, state.cos_window);


% calculate the response of the classifier
kf = fft2(dense_gauss_kernel(sigma, x, zp));
M = state.alphaf_num .* kf ./ state.alphaf_den;
M(isnan(M))=0;
response = real(ifft2(M)); clear M

% target location is at the maximum response
[row, col] = find(response == max(response(:)), 1);
pos = pos - floor(sz/2) + [row, col];

% AutoCorrection - Failures - out of screen
AutoCorrection = 0;
if (pos(2)<=2) || (pos(1)<=2)
    AutoCorrection = 1;
elseif (pos(2)>=cls-2)  || (pos(1)>=rows-2)
    AutoCorrection = 1;
end
if AutoCorrection==1
    pos = [rows/2, cls/2];
end

shift = 2;%min([target_sz*0.085]);
rect_bbox = [pos(2)-target_sz(2)/2,  pos(1)-target_sz(1)/2, target_sz(2) target_sz(1)];
rect_bbox_left = [pos(2)-shift-(target_sz(2))/2,  pos(1)-(target_sz(1))/2, target_sz(2), target_sz(1)];
rect_bbox_right = [pos(2)+shift-(target_sz(2))/2,  pos(1)-(target_sz(1))/2, target_sz(2), target_sz(1)];
rect_bbox_up = [pos(2)-(target_sz(2))/2,  pos(1)-shift-(target_sz(1))/2, target_sz(2), target_sz(1)];
rect_bbox_down = [pos(2)-(target_sz(2))/2,  pos(1)+shift-(target_sz(1))/2, target_sz(2), target_sz(1)];

sub_patch = imcrop(im,rect_bbox);
sub_patch_l = imcrop(im,rect_bbox_left);
sub_patch_r = imcrop(im,rect_bbox_right);
sub_patch_u = imcrop(im,rect_bbox_up);
sub_patch_d = imcrop(im,rect_bbox_down);
if ~isempty(sub_patch) && ~isempty(sub_patch_l) && ~isempty(sub_patch_r) && ~isempty(sub_patch_u) && ~isempty(sub_patch_d)
    
    curr_err = [];
    I = imresize(double(sub_patch)/255,[state.wsize(1),state.wsize(2)]);
    % extract 8 x 8 patches
    X=im2col(I,[4 4],'sliding');
    X=X-repmat(mean(X),[size(X,1) 1]);
    X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
    curr_err(1) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
    I = imresize(double(sub_patch_l)/255,[state.wsize(1),state.wsize(2)]);
    % extract 8 x 8 patches
    X=im2col(I,[4 4],'sliding');
    X=X-repmat(mean(X),[size(X,1) 1]);
    X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
    curr_err(2) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
    I = imresize(double(sub_patch_r)/255,[state.wsize(1),state.wsize(2)]);
    % extract 8 x 8 patches
    X=im2col(I,[4 4],'sliding');
    X=X-repmat(mean(X),[size(X,1) 1]);
    X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
    curr_err(3) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
    I = imresize(double(sub_patch_u)/255,[state.wsize(1),state.wsize(2)]);
    % extract 8 x 8 patches
    X=im2col(I,[4 4],'sliding');
    X=X-repmat(mean(X),[size(X,1) 1]);
    X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
    curr_err(4) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
    I = imresize(double(sub_patch_d)/255,[state.wsize(1),state.wsize(2)]);
    % extract 8 x 8 patches
    X=im2col(I,[4 4],'sliding');
    X=X-repmat(mean(X),[size(X,1) 1]);
    X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
    curr_err(5) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
    [~,id] = min([curr_err]);
    switch id
        case 2
            % Update box target size
            pos(2) = pos(2) - shift;
        case 3
            pos(2) = pos(2) + shift;
        case 4
            pos(1) = pos(1) - shift;
        case 5
            pos(1) = pos(1) + shift;
    end
    scale=1;
    rect_bbox = [pos(2)-target_sz(2)/2,  pos(1)-target_sz(1)/2, target_sz(2) target_sz(1)];
    sub_patch = imcrop(im,rect_bbox);
    big_size = target_sz + scale;
    rect_bbox_b = [pos(2)-big_size(2)/2,  pos(1)-big_size(1)/2, big_size(2) big_size(1)];
    sub_patch_b = imcrop(im,rect_bbox_b);
    small_size = target_sz - scale;
    rect_bbox_s = [pos(2)-small_size(2)/2,  pos(1)-small_size(1)/2, small_size(2) small_size(1)];
    sub_patch_s = imcrop(im,rect_bbox_s);
    if ~isempty(sub_patch_s) && ~isempty(sub_patch_b)
        curr_err = [];
        I = imresize(double(sub_patch)/255,[state.wsize(1),state.wsize(2)]);
        % extract 8 x 8 patches
        X=im2col(I,[4 4],'sliding');
        X=X-repmat(mean(X),[size(X,1) 1]);
        X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
        curr_err(1) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
        I = imresize(double(sub_patch_b)/255,[state.wsize(1),state.wsize(2)]);
        % extract 8 x 8 patches
        X=im2col(I,[4 4],'sliding');
        X=X-repmat(mean(X),[size(X,1) 1]);
        X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
        curr_err(2) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
        I = imresize(double(sub_patch_s)/255,[state.wsize(1),state.wsize(2)]);
        % extract 8 x 8 patches
        X=im2col(I,[4 4],'sliding');
        X=X-repmat(mean(X),[size(X,1) 1]);
        X=X./ repmat(sqrt(sum(X.^2)),[size(X,1) 1]); X(isnan(X))=0;
        curr_err(3) = mean(0.5*sum((X-state.D*state.alpha).^2)+state.lambda*sum(abs(state.alpha)));
        [~,id] = min(curr_err);
        switch id
            case 2
                
                % Update box target size
                if max(big_size) <  1.5*max([state.wsize(1),state.wsize(2)])
                    target_sz = big_size;
                end
            case 3
                % Update box target size
                if min(small_size) > 10 && min(small_size) > 0.4*min([state.wsize(1),state.wsize(2)])
                    target_sz = small_size;
                end
        end
    end
end

% extract the feature map of the local image patch to train the classifer
[xo_npca, xo_pca, xo_pca_neg ] = get_subwindow_fs(im, pos, sz, non_compressed_features, compressed_features, w2c);

% update the appearance
state.z_npca = (1 - learning_rate) * state.z_npca + learning_rate * xo_npca;
state.z_pca = (1 - learning_rate) * state.z_pca + learning_rate * xo_pca;

% FEATURE SELECTION
[ feature_ranking ] = supINFFS( xo_pca_neg, xo_pca );

% compute the mean appearance
data_mean = mean(state.z_pca, 1);

% substract the mean from the appearance to get the data matrix
data_matrix = bsxfun(@minus, state.z_pca, data_mean);

data_matrix(:,feature_ranking(num_selected_dim+1:end))=0;
data_matrix = data_matrix(:,  feature_ranking);

% calculate the covariance matrix
cov_matrix = 1/(prod(sz) - 1) * (data_matrix' * data_matrix);

[pca_basis, pca_variances, ~] = svd([(1 - compression_learning_rate) * state.old_cov_matrix + compression_learning_rate * cov_matrix]);
% calculate the projection matrix as the first principal
% components and extract their corresponding variances
state.projection_matrix = pca_basis(:, 1:num_compressed_dim);
state.projection_variances = pca_variances(1:num_compressed_dim, 1:num_compressed_dim);

% update the old covariance matrix using the computed
% projection matrix and variances
state.old_cov_matrix = (1 - compression_learning_rate) * state.old_cov_matrix + compression_learning_rate * (state.projection_matrix * state.projection_variances * state.projection_matrix');

% project the features of the new appearance example using the new
% projection matrix
x = feature_projection(xo_npca, xo_pca, state.projection_matrix, state.cos_window);

% calculate the new classifier coefficients
kf = fft2(dense_gauss_kernel(sigma, x));
new_alphaf_num = state.yf .* kf;
new_alphaf_den = kf .* (kf + lambda);

% subsequent frames, update the model
state.alphaf_num = (1 - learning_rate) * state.alphaf_num + learning_rate * new_alphaf_num;
state.alphaf_den = (1 - learning_rate) * state.alphaf_den + learning_rate * new_alphaf_den;

location = [ [pos([2,1]) - (target_sz([2,1]))/2 ], target_sz([2,1])];
state.newsize = target_sz;
state.position = pos;

end
